local MT = MacroToolkit
local string, table, ipairs, pairs, type, math = string, table, ipairs, pairs, type, math
local strsplit, select, wipe, tonumber, tostring = strsplit, select, wipe, tonumber, tostring
local GetSpellInfo, GetItemInfo = GetSpellInfo, GetItemInfo
local _G = _G
local L = MT.L

function MT:FindShortest(cglobal)
	local shortest = string.rep("x", 99)
	local current
	for c = 1, 99 do
		current = _G[cglobal .. c]
		if current then if string.len(current) < string.len(shortest) then shortest = current end
		else break end
	end
	return shortest
end

function MT:BuildCommandList()
	local cglobal, command, shortest, digit, param, custom
	for k, v in pairs(_G) do
		if type(v) == "string" then
			if string.sub(k, 1, 6) == "SLASH_" then
				if not string.find(k, "STOPWATCH_PARAM_") then
					digit = string.find(k, "%d+$")
					cglobal = string.sub(k, 1, digit - 1)
					command = string.sub(v, 2)
					custom = nil
					shortest = string.sub(MT:FindShortest(cglobal), 2)
					param = MT.commandinfo[string.sub(cglobal, 7)] or 0
					for _, c in ipairs(MT.db.global.custom) do if c.n == command then custom = 1 end end
					MT.commands[command] = {shortest, param, nil, custom}
				end
			elseif string.find(k, "EMOTE%d+_CMD") then
				digit = string.find(k, "%d+$")
				if digit then
					cglobal = string.sub(k, 1, digit - 1)
					command = string.sub(v, 2)
					shortest = string.sub(MT:FindShortest(cglobal), 2)
					param = MT.emoteinfo[tonumber(string.sub(string.find(cglobal, "%d+")))] or 0
					MT.commands[command] = {shortest, param, 1}
				end
			end
		end
	end
end

local function trim(s) return string.match(s, "^%s*(.*%S)") or "" end
local function escape(s) return (s:gsub("[%-%.%+%[%]%(%)%$%^%%%?%*]","%%%1"):gsub("%z","%%z")) end

local function getLevenshtein(s, t)
	local s, t = tostring(s), tostring(t)
	local m, n, d = string.len(s), string.len(t), {}
	for i = 0, m do d[i] = {[0] = i} end
	for j = 1, n do d[0][j] = j end
	for i = 1, m do
		for j = 1, n do
			local cost = string.sub(s, i, i) == string.sub(t, j, j) and 0 or 1
			d[i][j] = math.min(d[i - 1][j] + 1, d[i][j - 1] + 1, d[i - 1][j - 1] + cost)
		end
	end
	return d[m][n]
end

local function findMatch(source, values)
	local diff, bestmatch = 99, ""
	local d
	for k, v in pairs(values) do
		d = getLevenshtein(source, k)
		if d < diff then
			diff = d
			bestmatch = k
		end
	end
	return bestmatch
end

local function validateCommandVerb(commandtext, parameters)
	local param = string.len(trim(parameters or "")) > 0
	local c, cc = "|c" .. MT.db.profile.defaultcolour, "|c" .. MT.db.profile.commandcolour
	local p = false
	commandtext = trim(commandtext)
	local msg = L["Invalid command"] .. ": " .. MT.slash .. cc .. commandtext
	for k, v in pairs(MT.commands) do
		if k == commandtext then
			if v[3] then cc = "|c" .. MT.db.profile.emotecolour end
			if v[2] == 5 then 
				msg = L["Command removed"] .. ": " .. MT.slash .. cc .. commandtext
				p = true
			elseif v[2] == 1 then
				if not param then 
					msg = L["Required parameter missing"] .. ": " .. MT.slash .. cc .. commandtext
					p = true
				else msg = nil end
			else msg = nil end
			break
		end
	end
	for _, s in ipairs(MT.scripts) do
		if commandtext == s[2] then
			cc = "|c" .. MT.db.profile.mtcolour
			break
		end
	end
	if not msg then c = cc
	else
		msg = msg .. "|r"
		if not p then msg = msg .. "\n      " .. L["did you mean"] .. ": " .. MT.slash .. findMatch(commandtext, MT.commands) .. "|r" end
	end
	return c, msg
end

local function isNumeric(args)
	local numeric = true
	local arg
	for _, a in ipairs(args) do
		if not tonumber(trim(a)) then
			numeric = nil
			arg = a
			break
		end
	end
	return numeric, arg
end

local function isAlphaNumeric(args)
	local alpha, arg
	for _, a in ipairs(args) do
		local s, e = string.find(a, "%w+")
		alpha = (s == 1 and e == string.len(a))
		if not alpha then
			arg = a
			break 
		end
	end
	return alpha, arg
end

local function isValid(args, opt)
	local valid, arg
	for _, a in ipairs(args) do
		valid = nil
		for _, o in ipairs(MT.optargs[opt]) do
			if o == trim(a) then
				valid = true
				break
			end
		end
		if not valid then
			arg = a
			break
		end
	end
	return valid, arg
end

local function validateParameters(parameters, commandtext)
	local c = "|c" .. MT.db.profile.stringcolour
	local err
	parameters = trim(parameters)
	commandtext = trim(commandtext)
	if string.sub(parameters, 1, 1) == "!" then parameters = string.sub(parameters, 2) end
	if commandtext == string.sub(_G.SLASH_SCRIPT1, 2) or commandtext == string.sub(_G.SLASH_SCRIPT2, 2) or 
		commandtext == string.sub(_G.SLASH_SCRIPT3, 2) or commandtext == string.sub(_G.SLASH_SCRIPT4, 2) or 
		commandtext == string.sub(_G.SLASH_CONSOLE1, 2) or commandtext == string.sub(_G.SLASH_CONSOLE2, 2) or 
		commandtext == string.sub(_G.SLASH_CLICK1, 2) or commandtext == string.sub(_G.SLASH_CLICK2, 2) then c = "|c" .. MT.db.profile.scriptcolour
	elseif commandtext == string.sub(_G.SLASH_TARGET1, 2) or commandtext == string.sub(_G.SLASH_TARGET2, 2) or 
		commandtext == string.sub(_G.SLASH_TARGET3, 2) or commandtext == string.sub(_G.SLASH_TARGET4, 2) then c = "|c" .. MT.db.profile.targetcolour
	elseif GetSpellInfo(parameters) then c = "|c" .. MT.db.profile.spellcolour
	elseif GetItemInfo(parameters) then c = "|c" .. MT.db.profile.itemcolour
	elseif isNumeric({parameters}) then c = "|c" .. MT.db.profile.stringcolour
	elseif MT.db.profile.unknown then err = L["Unknown parameter"] .. ": |c" .. MT.db.profile.stringcolour .. parameters .. "|r" end
	return c, err
end

local function validateCondition(condition, optionarguments)
	local c, cc, cond = "|c" .. MT.db.profile.defaultcolour, "|c" .. MT.db.profile.conditioncolour, true
	local msg = L["Invalid condition"] .. ": " .. cc .. condition
	local target, noa, valid, arg, no
	condition = trim(condition)
	if string.sub(condition,1,2) == "no" then
		condition = string.sub(condition,3)
		no = true
	end
	local s, e = string.find(condition, "target%s-=")
	if not s then s, e = string.find(condition, "@") end
	if s then
		if e then target = string.sub(condition, e + 1) end
		if not target then
			msg = L["Invalid target"]
			cond = false
		elseif string.find(trim(target),"[%p%s%c]") then
			msg = L["Invalid target"] .. ": |c" .. MT.db.profile.targetcolour .. trim(target)
			cond = false
		else msg = nil end
	else
		for k, v in pairs(MT.conditions) do
			if k == condition then
				if v > 0 and k ~= "group" and k ~= "mod" and k ~= "modifier" and k~= "pet" then
					if #optionarguments == 0 then 
						msg = L["Argument not optional"] .. ": " .. cc .. condition
						noa = true
						cond = false
					end
				end
				if #optionarguments > 0 then
					if v == 0 then
						msg = L["Invalid argument"] .. ": " .. cc .. condition
						cond = false
					elseif v == 1 then --validate numeric
						valid, arg = isNumeric(optionarguments)
						if not valid then
							msg = L["Arguments must be numeric"] .. ": " .. cc .. condition .. "|r - " .. arg
							cond = false
						else msg = nil end
					elseif v == 2 then --validate text
						if isNumeric(optionarguments) then
							msg = L["Arguments must not be numeric"] .. ": " .. cc .. condition
							cond = false
						else msg = nil end
					elseif v == 3 then --validate alphanumeric
						valid, arg = isAlphaNumeric(optionarguments)
						if not valid then
							msg = L["Arguments must be alphanumeric"] .. ": " .. cc .. condition .. "|r - " .. arg
							cond = false
						else msg = nil end
					elseif v > 3 then --validate group
						valid, arg = isValid(optionarguments, v)
						if not valid then
							msg = L["Invalid argument"] .. ": " .. cc .. condition .. "|r - " .. arg
							cond = false
						else msg = nil end
					end
				elseif not noa then msg = nil end
				break
			end
		end
	end
	if not msg then
		c = cc
	else
		msg = msg .. "|r"
		if cond then msg = msg .. "\n      " .. L["did you mean"] .. ": " .. cc .. (no and "no" or "") .. findMatch(condition, MT.conditions) .. "|r" end
	end
	return c, msg
end

local function replace(original, replacetext, replacewith, start)
	local offset = 0
	local inc
	local s, e = string.find(replacewith, "target%s-=")
	if s then e = s + 5 end
	if s and not string.find(replacewith, "|c" .. MT.db.profile.defaultcolour) then 
		local rw = string.sub(replacewith, 1, s - 1) .. "target|r"
		local s2, e2 = string.find(replacewith, "%s?=%s*", s)
		replacewith = rw .. string.sub(replacewith, s2, e2) .. "|c" .. MT.db.profile.targetcolour .. string.sub(replacewith, e2 + 1) .. "|r"
		inc = 14
	end
	s, e = string.find(replacewith, "@")
	if s and not string.find(replacewith, "|c" .. MT.db.profile.defaultcolour) then 
		replacewith = string.sub(replacewith, 1, s) .. "|c" .. MT.db.profile.targetcolour .. string.sub(replacewith, e + 1) .. "|r"
		inc = 12
	end
	s, e = string.find(original, escape(replacetext), start)
	local out = ""
	if s then out = string.sub(original, 1, s - 1) .. replacewith .. string.sub(original, e + 1) end
	return out, offset
end

function MT:Decolourise(macrotext)
	macrotext = string.gsub(macrotext, "|cff%x%x%x%x%x%x", "")
	macrotext = string.gsub(macrotext, "|r", "")
	return macrotext
end

local function matchedBrackets(macrotext)
	local _, left = string.gsub(macrotext, "%[", "")
	local _,right = string.gsub(macrotext, "%]", "")
	return (left - right == 0)
end

function MT:ShortenMacro(macrotext)
	local olen, nlen = 0, 0
	olen = string.len(macrotext)
	macrotext = trim(macrotext)
	if string.len(macrotext) == 0 then return end
	macrotext = string.gsub(macrotext, "%s-\n", "\n")
	macrotext = string.gsub(macrotext, "%s%s", " ")
	macrotext = string.gsub(macrotext, "%[%s+", "[")
	macrotext = string.gsub(macrotext, "%s-([%]=,;])", "%1")
	macrotext = string.gsub(macrotext, "([%[%]=,;])%s", "%1")
	macrotext = string.gsub(macrotext, "%s-/%s+", "/")
	macrotext = string.gsub(macrotext, "%s-;\n", "\n")
	macrotext = string.gsub(macrotext, "target=", "@")
	macrotext = string.gsub(macrotext, "modifier%s-", "mod")
	macrotext = string.gsub(macrotext, "mod%s-:%s+", "mod:")
	macrotext = string.gsub(macrotext, "group%s-:%s+", "group:")
	local s, e = string.find(macrotext, _G.SLASH_SHOWTOOLTIP1)
	if s then
		local s1, e1 = string.find(macrotext, _G.SLASH_SHOW1 .. "[%s\n]")
		if s1 then macrotext = string.gsub(macrotext, _G.SLASH_SHOW1 .. "[%s\n].-\n", "") end
	end
	local mlines = {strsplit("\n", macrotext .. "\n")}
	local spell, matched
	for _, l in ipairs(mlines) do
		local _, _, command, parameters = MT:ParseMacro(l)
		if command == string.sub(_G.SLASH_SHOW1, 2) or command == string.sub(_G.SLASH_SHOWTOOLTIP1, 2) then
			if parameters then spell = parameters end
		elseif command == string.sub(_G.SLASH_CAST1, 2) or command == string.sub(_G.SLASH_CAST2, 2) or
			command == string.sub(_G.SLASH_CAST3, 2) or command == string.sub(_G.SLASH_CAST4, 2) then
			if ((parameters or "") == spell) and not matched then
				macrotext = string.gsub(macrotext, "^" .. _G.SLASH_SHOW1 .. "%s.-\n", _G.SLASH_SHOW1 .. "\n")
				macrotext = string.gsub(macrotext, "^" .. _G.SLASH_SHOWTOOLTIP1 .. "%s.-\n", _G.SLASH_SHOWTOOLTIP1 ..  "\n")
				matched = true
			else matched = true end
		elseif command == string.sub(_G.SLASH_SCRIPT1, 2) or command == string.sub(_G.SLASH_SCRIPT2, 2) or 
			command == string.sub(_G.SLASH_SCRIPT3, 2) or command == string.sub(_G.SLASH_SCRIPT4, 2) then
			if not string.find(macrotext, "DEFAULT_CHAT_FRAME:AddMessage%(.-,%s-[%d,]%)") then
				macrotext = string.gsub(macrotext, "DEFAULT_CHAT_FRAME:AddMessage", "print")
			end
			macrotext = string.gsub(macrotext, "local%s+", "")
			macrotext = string.gsub(macrotext, "%s-([{}%(%)%-%+])", "%1")
			macrotext = string.gsub(macrotext, "([{}%(%)%-%+])%s", "%1")
			macrotext = string.gsub(macrotext, "table%.getn", "#")
			macrotext = string.gsub(macrotext, "getn", "#")
			macrotext = string.gsub(macrotext, "table.wipe", "wipe")
			macrotext = string.gsub(macrotext, "string%.format", "format")
			macrotext = string.gsub(macrotext, "string%.gsub", "gsub")
			macrotext = string.gsub(macrotext, "string%.reverse", "strrev")
			local rep = {"byte", "char", "find", "len", "lower", "match", "rep", "sub", "upper"}
			for _, r in ipairs(rep) do
				local bc1 = string.match(macrotext, "string%." .. r .. "%((.-%))")
				local bc2 = string.match(macrotext, "str" .. r .. "%((.-%))")
				local bc = bc1 and bc1 or bc2
				local p1, p2
				if bc then
					p1 = string.match(bc, "(.-)[,%)]") or ""
					if string.len(bc) ~= string.len(p1) + 1 then
						s, e = string.find(bc, ",")
						p2 = string.sub(bc, s + 1)
					end
					macrotext = string.gsub(macrotext, (bc2 and "str" or "string%.") .. r .. "%(.-%)", p1 .. ":" .. r .. "(" .. (p2 or ")"))
				end
			end
		end
		for c, d in pairs(MT.commands) do
			s, e = string.find(macrotext, MT.slash .. c .. "[%s\n]")
			if s then
				if s~= e then
					if string.len(string.sub(macrotext, s + 1, e - 1)) > string.len(d[1]) then
						macrotext = string.sub(macrotext, 1, s) .. d[1] .. string.sub(macrotext, e)
					end
				end
			end
		end
	end
	macrotext = string.gsub(macrotext, ";$", "")
	if MT.db.profile.replacemt then
		local srun, scon
		srun = MT:FindShortest("SLASH_SCRIPT")
		scon = MT:FindShortest("SLASH_CONSOLE")
		macrotext = macrotext .. "\n"
		for _, s in ipairs(MT.scripts) do
			if s[4] then
				macrotext = string.gsub(macrotext, ((s[4] == 1) and srun or scon) .. " " .. s[5] .. "\n", MT.slash .. (s[6] or s[2]) .. "\n")
			end
		end
		macrotext = trim(macrotext)
	end
	local nlen = olen - string.len(macrotext)
	return macrotext, nlen
end

function MT:ParseMacro(macrotext)
	if macrotext == "\n" then return macrotext, {} end
	if string.len(macrotext) == 0 then return nil, {} end
	local command_verb, options
	local command_objects, conditions, condition_phrases = {}, {}, {}
	local command_object, condition, condition_phrase, condition_string
	local option_arguments, parsed_text, errors = {}, {}, {}
	local parameters, option_word, option_argument, target
	local spos, schar, ss, se, pt, err, col, pos, out, vv, epos
	
	schar = string.sub(macrotext, 1, 1)
	if schar == "#" or schar == MT.slash then
		if not matchedBrackets(macrotext) then table.insert(errors, L["Unmatched brackets"]) end
		spos = string.find(macrotext, " ")
		if spos then command_verb = string.sub(macrotext, 2, spos - 1)
		else
			command_verb = string.sub(macrotext, 2)
			col, err = validateCommandVerb(command_verb)
			if err then table.insert(errors, err) end
			table.insert(parsed_text, {t = command_verb, c = col, s = 2})
			vv = true
		end
		if spos then options = string.sub(macrotext, spos + 1) end
		if command_verb == string.sub(_G.SLASH_SCRIPT1, 2) or command_verb == string.sub(_G.SLASH_SCRIPT2, 2) or 
			command_verb == string.sub(_G.SLASH_SCRIPT3, 2) or command_verb == string.sub(_G.SLASH_SCRIPT4, 2) then
			if spos then
				parameters = options
				col, err = validateParameters(parameters, command_verb)
				if err then table.insert(errors, err) end
				table.insert(parsed_text, {t = parameters, c = col, s = spos + 1})
				col, err = validateCommandVerb(command_verb, parameters)
				if err then table.insert(errors, err) end
				table.insert(parsed_text, {t = command_verb, c = col, s = 2})
			end
		elseif options then
			if string.find(options, ";") then command_objects = {strsplit(";", options)}
			else command_objects = {options} end
			for _, command_object in ipairs(command_objects)do
				wipe(conditions)
				ss, se = string.find(command_object, "%[.*%]")
				if ss then
					condition_string = string.sub(command_object, ss, se)
					for c in string.gmatch(condition_string, "%[(.-)%]") do
						pos = string.find(macrotext, escape(c))
						table.insert(conditions, {c = c, p = pos})
					end
					parameters = string.sub(command_object, se + 1)
				else parameters = command_object end
				if string.len(parameters or "") > 0 then
					col, err = validateParameters(parameters, command_verb)
					if err then table.insert(errors, err) end
					pos, epos = string.find(macrotext, escape(parameters), epos or 1)
					table.insert(parsed_text, {t = parameters, c = col, s = pos})
				end
				if not vv then
					col, err = validateCommandVerb(command_verb, parameters)
					if err then table.insert(errors, err) end
					table.insert(parsed_text, {t = command_verb, c = col, s = 2})
					vv = true
				end
				for _, c in ipairs(conditions) do
					local cps = {strsplit(",", c.c)}
					for _, condition_phrase in ipairs(cps) do
						spos = string.find(condition_phrase, ":")
						wipe(option_arguments)
						if spos then
							option_arguments = {strsplit("/", string.sub(condition_phrase, spos + 1))}
							condition = string.sub(condition_phrase, 1, spos - 1)
						else condition = condition_phrase end
						col, err = validateCondition(condition, option_arguments)
						if err then table.insert(errors, err) end
						pos = string.find(macrotext, escape(condition), c.p)
						table.insert(parsed_text, {t = condition, c = col, s = pos})
					end
				end
			end
		end
	else
		table.insert(errors, L["Not a macro command"] .. ": " .. "|c" .. MT.db.profile.errorcolour .. macrotext .. "|r")
	end
	
	table.sort(parsed_text, function(a,b) return a.s < b.s end)
	out = macrotext

	local offset, inc = 0, 0
	for _, term in ipairs(parsed_text) do
		out, inc = replace(out, term.t, term.c .. term.t .. "|r", term.s + offset)
		offset = offset + 12 + inc
	end

	return out, errors, command_verb, parameters
end
		